#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################
import torch

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
from myutils import create_single_lead_csv
from dataset import *
from model import *
import torch.nn as nn
from time import time
from myutils import *
import torch
#import matplotlib.pyplot as plt
from sklearn.metrics import classification_report

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################
def finetune_segmentation_network(model_folder, locations_info, fold, verbose):
    # Extract the features and labels.
    if verbose >= 1:
        print('Set dataset.')
    # ============================================================================================================
    # ============================================================================================================
    # create single dataset
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelDataset1D(train_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    valid_ = MurmurSingleChannelDataset1D(valid_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights),
                                                                   len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler=train_sampler,
                                               drop_last=True)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=1, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model.to(device)
    if device == 'cpu':
        state_dict = torch.load('segmentation_pretrain_best_model.pt', map_location='cpu')
    else:
        state_dict = torch.load('segmentation_pretrain_best_model.pt')
    model.load_state_dict(state_dict)

    # set training parameters
    lr = 1e-4
    lr_step = False
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    if lr_step:
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 30, 50, 80, 120, 150], gamma=0.5)
    loss_class = nn.CrossEntropyLoss()
    loss_seg = DiceLoss()

    # begin training
    if verbose >= 1:
        print('Begin training.')
    epochs = 15
    cur_valid_loss = np.inf
    for epoch in range(1, epochs + 1):
        # ===================================================
        # Train
        # ===================================================
        model.train()
        epoch_loss = []
        epoch_segloss = []
        epoch_classloss = []
        epoch_step_time = []
        for i, sample in enumerate(train_loader, 1):
            step_start_time = time()
            signal, mask, mask4, label = sample
            signal = signal.to(device, dtype=torch.float)
            mask4 = mask4.to(device, dtype=torch.float)
            mask = mask.to(device, dtype=torch.long)
            label = label.to(device, dtype=torch.long)
            predicts_seg, predicts_class = model(signal)
            ce_loss = torch.tensor(0)  # loss_class(predicts_class, label)
            dice_loss = loss_seg(predicts_seg, mask4)
            total_loss = dice_loss + 5 * ce_loss
            optimizer.zero_grad()
            total_loss.backward()  # (retain_graph=True)
            optimizer.step()
            epoch_loss.append(total_loss.item())
            epoch_classloss.append(ce_loss.item())
            epoch_segloss.append(dice_loss.item())
            if lr_step:
                scheduler.step()
            epoch_step_time.append(time() - step_start_time)
        epoch_info = 'Training Epoch %d/%d' % (epoch, epochs)
        time_info = '%.4f sec/step' % np.mean(epoch_step_time)
        loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(epoch_loss), np.mean(epoch_segloss),
                                                          np.mean(epoch_classloss))
        print(' - '.join((epoch_info, time_info, loss_info)), flush=True)

        # ===================================================
        # validation
        # ===================================================
        with torch.no_grad():
            model.eval()
            valid_epoch_loss = []
            epoch_segloss = []
            epoch_classloss = []
            epoch_step_time = []
            for i, sample in enumerate(valid_loader, 1):
                step_start_time = time()
                signal, mask, mask4, label = sample
                signal = signal.to(device, dtype=torch.float)
                mask4 = mask4.to(device, dtype=torch.float)
                mask = mask.to(device, dtype=torch.long)
                label = label.to(device, dtype=torch.long)
                predicts_seg, predicts_class = model(signal)
                ce_loss = torch.tensor(0)  # loss_class(predicts_class, label)
                dice_loss = loss_seg(predicts_seg, mask4)
                total_loss = dice_loss + 5 * ce_loss
                valid_epoch_loss.append(total_loss.item())
                epoch_classloss.append(ce_loss.item())
                epoch_segloss.append(dice_loss.item())
                if lr_step:
                    scheduler.step()
                epoch_step_time.append(time() - step_start_time)
            epoch_info = 'Validation Epoch %d/%d' % (epoch, epochs)
            time_info = '%.4f sec/step' % np.mean(epoch_step_time)
            valid_loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(valid_epoch_loss), np.mean(epoch_segloss),
                                                                    np.mean(epoch_classloss))
            print(' - '.join((epoch_info, time_info, valid_loss_info)), flush=True)
            with open(os.path.join(model_folder, 'segmentation_2022_validation_fold' + str(fold) + '.csv'), 'a') as f:
                f.write(str(epoch) + ',' + '%.4e' % (np.mean(epoch_loss)) + ',' + '%.4e' % (
                    np.mean(valid_epoch_loss)) + '\n')
        if np.mean(valid_epoch_loss) < cur_valid_loss:
            cur_valid_loss = np.mean(valid_epoch_loss)
            # save model
            torch.save(model.state_dict(), os.path.join(model_folder, 'segmentation_2022_best_model.pt'))
    if verbose >= 1:
        print('Done.')



def train_murmur_classifier(model_folder, locations_info, fold, verbose):
    # ============================================================================================================
    # Present Unknown Absent classifier
    # ============================================================================================================
    classes = ['Present', 'Unknown', 'Absent']
    num_classes = len(classes)

    # Extract the features and labels.
    # ============================================================================================================
    # ============================================================================================================
    # create single dataset
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelDataset1D(train_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    valid_ = MurmurSingleChannelDataset1D(valid_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights),
                                                                   len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler=train_sampler,
                                               drop_last=True)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=1, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    if device == 'cpu':
        state_dict = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'), map_location='cpu')
    else:
        state_dict = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'))
    model.load_state_dict(state_dict)
    model.classifier = nn.Sequential(
        nn.LayerNorm(64),
        nn.Linear(64, 16),
        nn.ReLU(),
        nn.Linear(16, num_classes)
    )
    model.to(device)
    softmax = nn.Softmax(dim=1)
    softmax.to(device)
    # set training parameters
    lr = 1e-4
    lr_step = True
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    if lr_step:
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 30, 50, 80, 120, 150], gamma=0.5)
    loss_class = nn.CrossEntropyLoss()
    loss_seg = DiceLoss()

    # begin training
    if verbose >= 1:
        print('Begin training.')
    epochs = 50
    cur_valid_loss = -np.inf
    for epoch in range(1, epochs + 1):
        # ===================================================
        # Train
        # ===================================================
        model.train()
        epoch_loss = []
        epoch_segloss = []
        epoch_classloss = []
        epoch_step_time = []

        for i, sample in enumerate(train_loader, 1):
            step_start_time = time()
            signal, mask, mask4, label = sample
            signal = signal.to(device, dtype=torch.float)
            mask4 = mask4.to(device, dtype=torch.float)
            mask = mask.to(device, dtype=torch.long)
            label = label.to(device, dtype=torch.long)
            predicts_seg, predicts_class = model(signal)
            ce_loss = loss_class(predicts_class, label)
            dice_loss = loss_seg(predicts_seg, mask4)
            total_loss = dice_loss + 10*ce_loss
            optimizer.zero_grad()
            total_loss.backward()  # (retain_graph=True)
            optimizer.step()
            epoch_loss.append(total_loss.item())
            epoch_classloss.append(ce_loss.item())
            epoch_segloss.append(dice_loss.item())
            if lr_step:
                scheduler.step()
            epoch_step_time.append(time() - step_start_time)
        epoch_info = 'Training Epoch %d/%d' % (epoch, epochs)
        time_info = '%.4f sec/step' % np.mean(epoch_step_time)
        loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(epoch_loss), np.mean(epoch_segloss),
                                                          np.mean(epoch_classloss))
        print(' - '.join((epoch_info, time_info, loss_info)), flush=True)

        # ===================================================
        # validation
        # ===================================================
        with torch.no_grad():
            model.eval()
            valid_epoch_loss = []
            epoch_segloss = []
            epoch_classloss = []
            epoch_step_time = []
            all_preds = []
            all_gts = []
            for i, sample in enumerate(valid_loader, 1):
                step_start_time = time()
                signal, mask, mask4, label = sample
                signal = signal.to(device, dtype=torch.float)
                mask4 = mask4.to(device, dtype=torch.float)
                mask = mask.to(device, dtype=torch.long)
                label = label.to(device, dtype=torch.long)
                predicts_seg, predicts_class = model(signal)
                ce_loss = loss_class(predicts_class, label)
                dice_loss = loss_seg(predicts_seg, mask4)
                total_loss = dice_loss + 10*ce_loss
                valid_epoch_loss.append(total_loss.item())
                epoch_classloss.append(ce_loss.item())
                epoch_segloss.append(dice_loss.item())

                preds = np.argmax(softmax(predicts_class).detach().cpu().numpy(),axis=1)
                gts = label.detach().cpu().numpy()
                all_preds.append(preds)
                all_gts.append(gts)

                if lr_step:
                    scheduler.step()
                epoch_step_time.append(time() - step_start_time)
            epoch_info = 'Validation Epoch %d/%d' % (epoch, epochs)
            time_info = '%.4f sec/step' % np.mean(epoch_step_time)
            valid_loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(valid_epoch_loss), np.mean(epoch_segloss),
                                                                    np.mean(epoch_classloss))
            print(' - '.join((epoch_info, time_info, valid_loss_info)), flush=True)
            all_gts = np.array(all_gts)
            all_preds = np.array(all_preds)
            #print(classification_report(np.array(all_preds), np.array(all_gts)))
            all_gts3 = np.zeros((len(all_gts), 3))
            for i in range(3):
                all_gts3[np.where(all_gts==i)[0], i] = 1
            all_preds3 = np.zeros((len(all_preds), 3))
            for i in range(3):
                all_preds3[np.where(all_preds==i)[0], i] = 1
            valid_score = compute_weighted_accuracy(all_gts3, all_preds3, classes)
            print(valid_score)
            with open(os.path.join(model_folder, 'murmur_train_validation_fold' + str(fold) + '.csv'), 'a') as f:
                f.write(str(epoch) + ',' + '%.4e' % (np.mean(epoch_loss)) + ',' + '%.4e' % (
                    np.mean(valid_epoch_loss)) + '\n')
        if valid_score > cur_valid_loss:
            cur_valid_loss = valid_score
            # save model
            torch.save(model.state_dict(), os.path.join(model_folder, 'murmur_best_model.pt'))
    if verbose >= 1:
        print('Done.')


def train_abnormal_model(model_folder, locations_info, fold, verbose):
    # ============================================================================================================
    # Normal abnormal classifier
    # ============================================================================================================
    classes = ['Abnormal', 'Normal']
    num_classes = len(classes)

    # Extract the features and labels.
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelDataset1D(train_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    valid_ = MurmurSingleChannelDataset1D(valid_df, resize_scale=32, signal_length=8192 * 2, decompose=True,
                                          dataset=2022, ori_mask=False)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights),
                                                                   len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler=train_sampler,
                                               drop_last=True)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=1, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    if device == 'cpu':
        state_dict = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'), map_location='cpu')
    else:
        state_dict = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'))
    model.load_state_dict(state_dict)
    model.to(device)
    softmax = nn.Softmax(dim=1)
    softmax.to(device)
    # set training parameters
    lr = 1e-4
    lr_step = False
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    if lr_step:
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 30, 50, 80, 120, 150], gamma=0.5)
    loss_class = nn.CrossEntropyLoss()
    loss_seg = DiceLoss()

    # begin training
    if verbose >= 1:
        print('Begin training.')
    epochs = 50
    cur_valid_loss = np.inf
    for epoch in range(1, epochs + 1):
        # ===================================================
        # Train
        # ===================================================
        model.train()
        epoch_loss = []
        epoch_segloss = []
        epoch_classloss = []
        epoch_step_time = []
        for i, sample in enumerate(train_loader, 1):
            step_start_time = time()
            signal, mask, mask4, label = sample
            signal = signal.to(device, dtype=torch.float)
            mask4 = mask4.to(device, dtype=torch.float)
            mask = mask.to(device, dtype=torch.long)
            label = label.to(device, dtype=torch.long)
            predicts_seg, predicts_class = model(signal)
            ce_loss = loss_class(predicts_class, label)
            dice_loss = loss_seg(predicts_seg, mask4)
            total_loss = dice_loss + 10 * ce_loss
            optimizer.zero_grad()
            total_loss.backward()  # (retain_graph=True)
            optimizer.step()
            epoch_loss.append(total_loss.item())
            epoch_classloss.append(ce_loss.item())
            epoch_segloss.append(dice_loss.item())
            if lr_step:
                scheduler.step()
            epoch_step_time.append(time() - step_start_time)
        epoch_info = 'Training Epoch %d/%d' % (epoch, epochs)
        time_info = '%.4f sec/step' % np.mean(epoch_step_time)
        loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(epoch_loss), np.mean(epoch_segloss),
                                                          np.mean(epoch_classloss))
        print(' - '.join((epoch_info, time_info, loss_info)), flush=True)

        # ===================================================
        # validation
        # ===================================================
        with torch.no_grad():
            model.eval()
            valid_epoch_loss = []
            epoch_segloss = []
            epoch_classloss = []
            epoch_step_time = []
            all_preds = []
            all_gts = []
            for i, sample in enumerate(valid_loader, 1):
                step_start_time = time()
                signal, mask, mask4, label = sample
                signal = signal.to(device, dtype=torch.float)
                mask4 = mask4.to(device, dtype=torch.float)
                mask = mask.to(device, dtype=torch.long)
                label = label.to(device, dtype=torch.long)
                predicts_seg, predicts_class = model(signal)
                ce_loss = loss_class(predicts_class, label)
                dice_loss = loss_seg(predicts_seg, mask4)
                total_loss = dice_loss + 10 * ce_loss
                valid_epoch_loss.append(total_loss.item())
                epoch_classloss.append(ce_loss.item())
                epoch_segloss.append(dice_loss.item())

                preds = np.argmax(softmax(predicts_class).detach().cpu().numpy(), axis=1)
                gts = label.detach().cpu().numpy()
                all_preds.append(preds)
                all_gts.append(gts)
                if lr_step:
                    scheduler.step()
                epoch_step_time.append(time() - step_start_time)
            epoch_info = 'Validation Epoch %d/%d' % (epoch, epochs)
            time_info = '%.4f sec/step' % np.mean(epoch_step_time)
            valid_loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(valid_epoch_loss), np.mean(epoch_segloss),
                                                                    np.mean(epoch_classloss))
            print(' - '.join((epoch_info, time_info, valid_loss_info)), flush=True)
            all_gts = np.array(all_gts)
            all_preds = np.array(all_preds)
            # print(classification_report(np.array(all_preds), np.array(all_gts)))
            all_gts3 = np.zeros((len(all_gts), 2))
            for i in range(2):
                all_gts3[np.where(all_gts == i)[0], i] = 1
            all_preds3 = np.zeros((len(all_preds), 2))
            for i in range(2):
                all_preds3[np.where(all_preds == i)[0], i] = 1
            valid_score = compute_cost(all_gts3, all_preds3, classes, classes)
            print(valid_score)
            with open(os.path.join(model_folder, 'outcome_train_validation_fold' + str(fold) + '.csv'), 'a') as f:
                f.write(str(epoch) + ',' + '%.4e' % (np.mean(epoch_loss)) + ',' + '%.4e' % (
                    np.mean(valid_epoch_loss)) + '\n')
        if valid_score < cur_valid_loss:
            cur_valid_loss = valid_score
            # save model
            torch.save(model.state_dict(), os.path.join(model_folder, 'outcome_best_model.pt'))
    if verbose >= 1:
        print('Done.')


# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)
    patient_info, murmur_locations_info, outcome_locations_info = create_single_lead_csv(patient_files, data_folder, model_folder, save=False)
    fold = 1
    # finetune segmentation
    finetune_segmentation_network(model_folder, murmur_locations_info, fold, verbose)
    # murmur classifier
    print('Train murmur')
    train_murmur_classifier(model_folder, murmur_locations_info, fold, verbose)
    # outcome classifier
    print('Train outcome')
    train_abnormal_model(model_folder, outcome_locations_info, fold, verbose)


# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    if device == 'cpu':
        state_dict_seg = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'), map_location='cpu')
        state_dict_murmur = torch.load(os.path.join(model_folder, 'murmur_best_model.pt'), map_location='cpu')
        state_dict_outcome = torch.load(os.path.join(model_folder, 'outcome_best_model.pt'), map_location='cpu')
    else:
        state_dict_seg = torch.load(os.path.join(model_folder, 'segmentation_2022_best_model.pt'))
        state_dict_murmur = torch.load(os.path.join(model_folder, 'murmur_best_model.pt'))
        state_dict_outcome = torch.load(os.path.join(model_folder, 'outcome_best_model.pt'))
    return {'state_dict_seg': state_dict_seg, 'state_dict_murmur': state_dict_murmur,
            'state_dict_outcome': state_dict_outcome, 'device': device}


# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    state_dict_seg = model['state_dict_seg']
    state_dict_murmur = model['state_dict_murmur']
    state_dict_outcome = model['state_dict_outcome']
    device = model['device']
    classes = ['Present', 'Unknown', 'Absent', 'Abnormal', 'Normal']

    seg_model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    seg_model.to(device)
    seg_model.load_state_dict(state_dict_seg)
    murmur_model = PCGModel(3, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    murmur_model.to(device)
    murmur_model.load_state_dict(state_dict_murmur)

    outcome_model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    outcome_model.to(device)
    outcome_model.load_state_dict(state_dict_outcome)

    softmax = nn.Softmax(dim=1)
    softmax.to(device)

    with torch.no_grad():
        seg_model.eval()
        murmur_model.eval()
        outcome_model.eval()
        locations = ['AV', 'MV', 'PV', 'TV']
        # get right order of recordings
        data_dict = load_all_seg_model_inputs(data, recordings)
        murmur_label_dict = {}
        murmur_prob_dict = {}
        outcome_label_dict = {}
        outcome_prob_dict = {}
        #print('New Patient')
        for l in locations:
            if data_dict[l] is not None:
                #print(len(data_dict[l]))
                sgi_ori, sig_down = data_dict[l]
                sig = torch.tensor(sig_down)
                sig = sig.to(device, dtype=torch.float)
                # segmentation
                pred_segs_prob, _ = seg_model(sig)
                pred_segs_prob = pred_segs_prob.detach().cpu().numpy()
                pred_segs_out = np.argmax(pred_segs_prob, axis=1)
                idx, idy = np.where(pred_segs_out==2)
                murmur_seg_prob = pred_segs_prob[idx, 2, idy]
                murmur_seg_mean = np.mean(murmur_seg_prob)
                # murmur + outcome detection
                _, murmur_out = murmur_model(sig)
                murmur_prob = np.mean(softmax(murmur_out).detach().cpu().numpy(), axis=0)
                _, outcome_out = outcome_model(sig)
                outcome_prob = np.mean(softmax(outcome_out).detach().cpu().numpy(), axis=0)
                murmur_prob[0] = (murmur_prob[0] + murmur_seg_mean)/2
                pos = np.argmax(murmur_prob)
                murmur_label = np.zeros(3)
                murmur_label[pos] = 1
                murmur_label_dict[l] = murmur_label
                murmur_prob_dict[l] = murmur_prob

                pos = np.argmax(outcome_prob)
                outcome_label = np.zeros(2)
                outcome_label[pos] = 1
                outcome_label_dict[l] = outcome_label
                outcome_prob_dict[l] = outcome_prob
        murmur_labels, murmur_probabilities = murmur_ensemble_output(murmur_label_dict, murmur_prob_dict)
        outcome_labels, outcome_probabilities = outcome_ensemble_output(outcome_label_dict, outcome_prob_dict)
    return classes, np.concatenate([murmur_labels, outcome_labels]), np.concatenate([murmur_probabilities, outcome_probabilities])

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################
# Compute a binary confusion matrix, where the columns are the expert labels and the rows are the classifier labels.
def compute_confusion_matrix(labels, outputs):
    assert(np.shape(labels)[0] == np.shape(outputs)[0])
    assert(all(value in (0, 1, True, False) for value in np.unique(labels)))
    assert(all(value in (0, 1, True, False) for value in np.unique(outputs)))

    num_patients = np.shape(labels)[0]
    num_label_classes = np.shape(labels)[1]
    num_output_classes = np.shape(outputs)[1]

    A = np.zeros((num_output_classes, num_label_classes))
    for k in range(num_patients):
        for i in range(num_output_classes):
            for j in range(num_label_classes):
                if outputs[k, i] == 1 and labels[k, j] == 1:
                    A[i, j] += 1

    return A


def compute_weighted_accuracy(labels, outputs, classes):
    # Define constants.
    if classes == ['Present', 'Unknown', 'Absent']:
        weights = np.array([[5, 3, 1], [5, 3, 1], [5, 3, 1]])
    elif classes == ['Abnormal', 'Normal']:
        weights = np.array([[5, 1], [5, 1]])
    else:
        raise NotImplementedError('Weighted accuracy undefined for classes {}'.format(', '.join(classes)))

    # Compute confusion matrix.
    assert(np.shape(labels) == np.shape(outputs))
    A = compute_confusion_matrix(labels, outputs)

    # Multiply the confusion matrix by the weight matrix.
    assert(np.shape(A) == np.shape(weights))
    B = weights * A

    # Compute weighted_accuracy.
    if np.sum(B) > 0:
        weighted_accuracy = np.trace(B) / np.sum(B)
    else:
        weighted_accuracy = float('nan')

    return weighted_accuracy

# Define total cost for algorithmic prescreening of m patients.
def cost_algorithm(m):
    return 10*m

# Define total cost for expert screening of m patients out of a total of n total patients.
def cost_expert(m, n):
    return (25 + 397*(m/n) -1718*(m/n)**2 + 11296*(m/n)**4) * n

# Define total cost for treatment of m patients.
def cost_treatment(m):
    return 10000*m

# Define total cost for missed/late treatement of m patients.
def cost_error(m):
    return 50000*m

# Compute Challenge cost metric.
def compute_cost(labels, outputs, label_classes, output_classes):
    # Define positive and negative classes for referral and treatment.
    positive_classes = ['Present', 'Unknown', 'Abnormal']
    negative_classes = ['Absent', 'Normal']

    # Compute confusion matrix.
    A = compute_confusion_matrix(labels, outputs)

    # Identify positive and negative classes for referral.
    idx_label_positive = [i for i, x in enumerate(label_classes) if x in positive_classes]
    idx_label_negative = [i for i, x in enumerate(label_classes) if x in negative_classes]
    idx_output_positive = [i for i, x in enumerate(output_classes) if x in positive_classes]
    idx_output_negative = [i for i, x in enumerate(output_classes) if x in negative_classes]

    # Identify true positives, false positives, false negatives, and true negatives.
    tp = np.sum(A[np.ix_(idx_output_positive, idx_label_positive)])
    fp = np.sum(A[np.ix_(idx_output_positive, idx_label_negative)])
    fn = np.sum(A[np.ix_(idx_output_negative, idx_label_positive)])
    tn = np.sum(A[np.ix_(idx_output_negative, idx_label_negative)])
    total_patients = tp + fp + fn + tn

    # Compute total cost for all patients.
    total_cost = cost_algorithm(total_patients) \
        + cost_expert(tp + fp, total_patients) \
        + cost_treatment(tp) \
        + cost_error(fn)

    # Compute mean cost per patient.
    if total_patients > 0:
        mean_cost = total_cost / total_patients
    else:
        mean_cost = float('nan')

    return mean_cost


def max_temporal_modelling(s, start=0):
    s = s - 1
    new_s = np.zeros(len(s))
    new_s[:start + 1] = s[:start + 1]
    for i in range(start + 1, len(s)):
        if (new_s[i - 1] + 1) % 4 == s[i]:
            new_s[i] = s[i]
        else:
            new_s[i] = new_s[i - 1]
    return np.array(new_s) + 1


def murmur_ensemble_output(label_dict, prob_dict):
    if len(label_dict) ==1:
        return list(label_dict.values())[0], list(prob_dict.values())[0]
    elif len(label_dict) > 1:
        # first prob setting
        prob = np.array(list(prob_dict.values()))
        new_prob = np.zeros((3))
        new_prob[0] = np.mean(prob[:, 0])
        new_prob[1] = np.mean(prob[:, 1])
        new_prob[2] = 1 - new_prob[0] - new_prob[1]
        # then label setting
        new_label = np.zeros(3)
        new_label[np.argmax(new_prob)] = 1
        return new_label, new_prob


def outcome_ensemble_output(label_dict, prob_dict):
    if len(label_dict) == 1:
        return list(label_dict.values())[0], list(prob_dict.values())[0]
    elif len(label_dict) > 1:
        # first prob setting
        prob = np.array(list(prob_dict.values()))
        new_prob = np.zeros((2))
        new_prob[0] = np.max(prob[:, 0])
        new_prob[1] = 1 - new_prob[0]
        # then label setting
        new_label = np.zeros(2)
        new_label[np.argmax(new_prob)] = 1
        return new_label, new_prob


def preprocess(record, signal_length=8192 * 2):
    # downsample
    drecordings = downsample(record, factor=4)
    # filtering
    frecordings = filtering(drecordings, low=25, high=400, order=4, sampling_rate=1000)
    # remove spikes
    srecordings = remove_spikes(frecordings, terminal=500, sampling=1000)
    # normalize
    nrecordings = normalize_signal(srecordings)
    # decomposition
    original = decompose_into_n_signals(nrecordings)
    original.append(nrecordings)
    # random cut records
    out_sig = []
    for i in range(len(original)):
        leng = len(original[i])
        cur = original[i]
        if leng > signal_length:
            s = np.random.randint(0, leng - signal_length + 1, 1)[0]
            cur = cur[s:s + signal_length]
        else:
            cur = np.pad(cur, (0, signal_length - leng))
        out_sig.append(cur)
    return np.array(out_sig)


def load_all_records(data, recordings, transform=True):
    record_dict = {}
    recording_locations = ['AV', 'MV', 'PV', 'TV', 'Phc']
    for r in recording_locations:
        record_dict[r] = None
    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations + 1]
    for i in range(num_locations):
        recording = recordings[i]
        entries = recording_information[i].split(' ')
        if record_dict[entries[0]] is None:
            record_dict[entries[0]] = recording
    out_data = {}
    for r in recording_locations:
        out_data[r] = None
    if transform:
        for key in record_dict.keys():
            sig = record_dict[key]
            if sig is None:
                out_data[key] = None
            else:
                out_data[key] = preprocess(sig)
    return out_data


def load_all_seg_model_inputs(data, recordings):
    record_dict = {}
    recording_locations = ['AV', 'MV', 'PV', 'TV', 'Phc']
    for r in recording_locations:
        record_dict[r] = None
    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations + 1]
    for i in range(num_locations):
        recording = recordings[i]
        entries = recording_information[i].split(' ')
        if record_dict[entries[0]] is None:
            record_dict[entries[0]] = recording
    out_data = {}
    for r in recording_locations:
        out_data[r] = None
    for key in record_dict.keys():
        sig = record_dict[key]
        if sig is None:
            out_data[key] = None
        else:
            out_data[key] = preprocess1D(sig)
    return out_data


def preprocess1D(record, signal_length=8192 * 2, resize_scale=32, fre=4000):
    # cut original signal into several pieces
    # return original frequency signal + down-sampled signal
    all_len = len(record)
    pieces5_ori = []
    pieces5_down = []
    # random cut signal
    num_slice = int(all_len//signal_length)
    if num_slice==0:
        num_slice += 1
    for i in range(num_slice):
        if all_len > signal_length:
            s = signal_length * i
            new_r = record[s:s + signal_length].copy()
        else:
            new_r = np.pad(record, (0, signal_length - all_len))
        # filtering
        frecordings = filtering(new_r, low=25, high=400, order=4, sampling_rate=fre)
        # remove spikes
        srecordings = remove_spikes(frecordings, terminal=500, sampling=fre)
        # normalize
        nrecordings = normalize_signal(srecordings)
        # decomposition
        decomp_records = decompose_into_n_signals(nrecordings)
        decomp_records = [np.abs(hilbert(r)) for r in decomp_records]
        decomp_records.append(np.abs(hilbert(nrecordings)))
        decomp_records = np.array(decomp_records)
        pieces5_ori.append(nrecordings)
        # downsample to 125Hz
        resize_records = [decimate(sig, resize_scale, ftype='fir') for sig in decomp_records]
        pieces5_down.append(resize_records)
    return [np.array(pieces5_ori), np.array(pieces5_down)]
